"""
Views for handling REST endpoints related to Engagements analytics.
"""

from datetime import date
from logging import getLogger

from edx_rbac.decorators import permission_required
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.viewsets import ViewSet

from django.http import StreamingHttpResponse

from enterprise_data.admin_analytics.constants import ResponseType
from enterprise_data.admin_analytics.database.tables import FactEngagementAdminDashTable, FactEnrollmentAdminDashTable
from enterprise_data.api.v1.serializers import AdvanceAnalyticsQueryParamSerializer
from enterprise_data.api.v1.views.base import AnalyticsPaginationMixin
from enterprise_data.renderers import IndividualEngagementsCSVRenderer
from enterprise_data.utils import timer

LOGGER = getLogger(__name__)


class AdvanceAnalyticsEngagementView(AnalyticsPaginationMixin, ViewSet):
    """
    View to handle requests for enterprise engagement data.

    Here is the list of URLs that are handled by this view:
    1. `enterprise_data_api_v1.enterprise-learner-engagement-list`: Get individual engagement data.
    2. `enterprise_data_api_v1.enterprise-learner-engagement-stats`: Get engagement stats data.
    """
    authentication_classes = (JwtAuthentication,)
    http_method_names = ('get', )

    @permission_required('can_access_enterprise', fn=lambda request, enterprise_uuid: enterprise_uuid)
    def list(self, request, enterprise_uuid):
        """
        Get individual engagements data for the enterprise.
        """
        # Remove hyphens from the UUID
        enterprise_uuid = enterprise_uuid.replace('-', '')

        serializer = AdvanceAnalyticsQueryParamSerializer(data=request.GET)
        serializer.is_valid(raise_exception=True)
        min_enrollment_date, _ = FactEnrollmentAdminDashTable().get_enrollment_date_range(
            enterprise_uuid,
        )

        # get values from query params or use default values
        start_date = serializer.data.get('start_date', min_enrollment_date)
        end_date = serializer.data.get('end_date', date.today())
        page = serializer.data.get('page', 1)
        page_size = serializer.data.get('page_size', 100)
        engagements = FactEngagementAdminDashTable().get_all_engagements(
            enterprise_customer_uuid=enterprise_uuid,
            start_date=start_date,
            end_date=end_date,
            limit=page_size,
            offset=(page - 1) * page_size,
        )
        total_count = FactEngagementAdminDashTable().get_engagement_count(
            enterprise_customer_uuid=enterprise_uuid,
            start_date=start_date,
            end_date=end_date,
        )
        response_type = request.query_params.get('response_type', ResponseType.JSON.value)

        LOGGER.info(
            "Individual engagements data requested for enterprise [%s] from [%s] to [%s]",
            enterprise_uuid,
            start_date,
            end_date,
        )

        if response_type == ResponseType.CSV.value:
            filename = f"""Individual Engagements, {start_date} - {end_date}.csv"""

            return StreamingHttpResponse(
                IndividualEngagementsCSVRenderer().render(self._stream_serialized_data(
                    enterprise_uuid, start_date, end_date, total_count
                )),
                content_type="text/csv",
                headers={"Content-Disposition": f'attachment; filename="{filename}"'},
            )

        return self.get_paginated_response(
            request=request,
            records=engagements,
            page=page,
            page_size=page_size,
            total_count=total_count,
        )

    @staticmethod
    def _stream_serialized_data(enterprise_uuid, start_date, end_date, total_count, page_size=50000):
        """
        Stream the serialized data.
        """
        offset = 0
        while offset < total_count:
            engagements = FactEngagementAdminDashTable().get_all_engagements(
                enterprise_customer_uuid=enterprise_uuid,
                start_date=start_date,
                end_date=end_date,
                limit=page_size,
                offset=offset,
            )
            yield from engagements
            offset += page_size

    @permission_required('can_access_enterprise', fn=lambda request, enterprise_uuid: enterprise_uuid)
    @action(detail=False, methods=['get'], name='Enterprise engagements data for charts', url_path='stats')
    def stats(self, request, enterprise_uuid):
        """
        Get data to populate enterprise engagement charts.

        Here is the list of the charts and their corresponding data:
        1. `engagement_over_time`: This will show time series data of engagements over time.
        2. `top_courses_by_engagement`: This will show the top courses by engagements.
        3. `top_subjects_by_engagement`: This will show the top subjects by engagements.
        """
        # Remove hyphens from the UUID
        enterprise_uuid = enterprise_uuid.replace('-', '')

        serializer = AdvanceAnalyticsQueryParamSerializer(data=request.GET)
        serializer.is_valid(raise_exception=True)

        min_enrollment_date, _ = FactEnrollmentAdminDashTable().get_enrollment_date_range(
            enterprise_uuid,
        )
        # get values from query params or use default
        start_date = serializer.data.get('start_date', min_enrollment_date)
        end_date = serializer.data.get('end_date', date.today())
        with timer('construct_engagement_all_stats'):
            data = {
                'engagement_over_time': FactEngagementAdminDashTable().get_engagement_time_series_data(
                    enterprise_uuid, start_date, end_date
                ),
                'top_courses_by_engagement': FactEngagementAdminDashTable().get_top_courses_by_engagement(
                    enterprise_uuid, start_date, end_date,
                ),
                'top_subjects_by_engagement': FactEngagementAdminDashTable().get_top_subjects_by_engagement(
                    enterprise_uuid, start_date, end_date,
                ),
            }
        return Response(data)



"""
Tests for clients in enterprise_reporting.
"""
from datetime import datetime, timedelta
from unittest.mock import Mock, patch
from urllib.parse import urljoin

import responses

from django.conf import settings
from django.test import TestCase

from enterprise_reporting.clients.enterprise import EnterpriseAPIClient, EnterpriseCatalogAPIClient


class TestEnterpriseAPIClient(TestCase):
    """
    Test Enterprise API client used to connect to the Enterprise API.
    """

    def setUp(self):
        self.enterprise_customer_uuid = 'test-enterprise-customer-uuid'
        self.api_response = {
            'results': [{
                'enterprise_name': 'Test Enterprise',
                'enterprise_id': 'test-id'
            }]
        }
        self.mocked_get_endpoint = Mock(return_value=self.api_response)
        self.client = EnterpriseAPIClient(  # pylint: disable=attribute-defined-outside-init
            settings.BACKEND_SERVICE_EDX_OAUTH2_KEY,
            settings.BACKEND_SERVICE_EDX_OAUTH2_SECRET,
        )
        self.client.API_BASE_URL = 'http://localhost-test:8000/'
        super().setUp()

    @responses.activate
    @patch('enterprise_reporting.clients.get_oauth_access_token')
    def test_get_all_enterprise_reporting_configs(self, mock_get_oauth_access_token):
        mock_get_oauth_access_token.return_value = ['test_access_token', datetime.now() + timedelta(minutes=60)]
        url = urljoin(self.client.API_BASE_URL + '/', self.client.ENTERPRISE_REPORTING_ENDPOINT)
        responses.add(
            responses.GET,
            url,
            json=self.mocked_get_endpoint(),
            status=200,
            content_type='application/json'
        )
        results = self.client.get_all_enterprise_reporting_configs()
        assert results['results'] == self.api_response['results']

    @responses.activate
    @patch('enterprise_reporting.clients.get_oauth_access_token')
    def test_get_enterprise_reporting_configs(self, mock_get_oauth_access_token):
        mock_get_oauth_access_token.return_value = ['test_access_token', datetime.now() + timedelta(minutes=60)]
        url = urljoin(self.client.API_BASE_URL + '/', self.client.ENTERPRISE_REPORTING_ENDPOINT)
        responses.add(
            responses.GET,
            url,
            json=self.mocked_get_endpoint(),
            status=200,
            content_type='application/json'
        )
        results = self.client.get_enterprise_reporting_configs(self.enterprise_customer_uuid)
        assert results['results'] == self.api_response['results']


class TestEnterpriseCatalogAPIClient(TestCase):
    """
    Test Enterprise Catalog API client used to connect to the Enterprise API.
    """

    def setUp(self):
        self.enterprise_customer_uuid = 'test-enterprise-customer-uuid'
        self.program_uuid = 'test-program_uuid'
        self.client = EnterpriseCatalogAPIClient(  # pylint: disable=attribute-defined-outside-init
            settings.BACKEND_SERVICE_EDX_OAUTH2_KEY,
            settings.BACKEND_SERVICE_EDX_OAUTH2_SECRET,
        )
        self.client.API_BASE_URL = 'http://localhost-test:8000/'
        super().setUp()

    @responses.activate
    @patch('enterprise_reporting.clients.get_oauth_access_token')
    def test_get_customer_catalogs(self, mock_get_oauth_access_token):
        mock_get_oauth_access_token.return_value = ['test_access_token', datetime.now() + timedelta(minutes=60)]
        api_response = {
            'results': [{
                'enterprise_name': 'Test Enterprise',
                'enterprise_id': 'test-id'
            }]
        }
        mocked_get_endpoint = Mock(return_value=api_response)
        url = urljoin(self.client.API_BASE_URL, self.client.ENTERPRISE_CATALOGS_ENDPOINT)
        url = f'{url}?enterprise_customer={self.enterprise_customer_uuid}&page_size=1000'
        responses.add(
            responses.GET,
            url,
            json=mocked_get_endpoint(),
            status=200,
            content_type='application/json'
        )
        results = self.client.get_customer_catalogs(self.enterprise_customer_uuid)
        assert results['results'] == api_response['results']

    @responses.activate
    @patch('enterprise_reporting.clients.get_oauth_access_token')
    def test_get_content_metadata(self, mock_get_oauth_access_token):
        mock_get_oauth_access_token.return_value = ['test_access_token', datetime.now() + timedelta(minutes=60)]
        metadata_api_response = {
            'results': [{
                'uuid': self.program_uuid,
                'content_type': 'program',
            }]
        }
        mocked_get_metadata_endpoint = Mock(return_value=metadata_api_response)
        url_path = self.client.GET_CONTENT_METADATA_ENDPOINT.format(self.enterprise_customer_uuid)
        url = urljoin(
            self.client.API_BASE_URL,
            f'{url_path}?page_size=1000'
        )
        responses.add(
            responses.GET,
            url,
            json=mocked_get_metadata_endpoint(),
            status=200,
            content_type='application/json'
        )
        request_catalogs = {
            'results': [{
                'uuid': self.enterprise_customer_uuid,
            }]
        }
        results = self.client.get_content_metadata(request_catalogs)
        assert results[0]['uuid'] == metadata_api_response['results'][0]['uuid']
        assert results[0]['content_type'] == metadata_api_response['results'][0]['content_type']

    @responses.activate
    @patch('enterprise_reporting.clients.get_oauth_access_token')
    def test_accepted_subject_in_content_metadata(self, mock_get_oauth_access_token):
        mock_get_oauth_access_token.return_value = ['test_access_token', datetime.now() + timedelta(minutes=60)]
        metadata_api_response = {
            'results': [{
                'key': 'test_key',
                'uuid': 'test_course_uuid',
                'content_type': 'course',
                'subjects': [
                    {
                        'name': 'Business & Management',
                    },
                    'Communication'
                ],
            }]
        }
        mocked_get_metadata_endpoint = Mock(return_value=metadata_api_response)
        url_path = self.client.GET_CONTENT_METADATA_ENDPOINT.format(self.enterprise_customer_uuid)
        url = urljoin(
            self.client.API_BASE_URL,
            f'{url_path}?page_size=1000'
        )
        responses.add(
            responses.GET,
            url,
            json=mocked_get_metadata_endpoint(),
            status=200,
            content_type='application/json'
        )
        request_catalogs = {
            'results': [{
                'uuid': self.enterprise_customer_uuid,
            }]
        }
        results = self.client.get_content_metadata(request_catalogs)
        assert results[0]['uuid'] == metadata_api_response['results'][0]['uuid']

    @responses.activate
    @patch('enterprise_reporting.clients.get_oauth_access_token')
    def test_rejected_subject_format_in_content_metadata(self, mock_get_oauth_access_token):
        mock_get_oauth_access_token.return_value = ['test_access_token', datetime.now() + timedelta(minutes=60)]
        metadata_api_response = {
            'results': [{
                'key': 'test_key',
                'uuid': 'test_course_uuid',
                'content_type': 'course',
                'subjects': [
                    [
                        'Communication'
                    ]
                ],
            }]
        }
        mocked_get_metadata_endpoint = Mock(return_value=metadata_api_response)
        url_path = self.client.GET_CONTENT_METADATA_ENDPOINT.format(self.enterprise_customer_uuid)
        url = urljoin(
            self.client.API_BASE_URL,
            f'{url_path}?page_size=1000'
        )
        responses.add(
            responses.GET,
            url,
            json=mocked_get_metadata_endpoint(),
            status=200,
            content_type='application/json'
        )
        request_catalogs = {
            'results': [{
                'uuid': self.enterprise_customer_uuid,
            }]
        }
        expected_error_message = "Subject is not a string or a dictionary: ['Communication']"
        try:
            results = self.client.get_content_metadata(request_catalogs)
        except Exception as e:
            self.assertEqual(str(e), expected_error_message)



"""
Test utilities.
"""


import datetime
import os
import tempfile
import unittest
from collections import OrderedDict

import ddt
import pgpy
import pytz
from pgpy.constants import CompressionAlgorithm, HashAlgorithm, KeyFlags, PubKeyAlgorithm, SymmetricKeyAlgorithm
from pgpy.errors import PGPError

from enterprise_reporting import utils

from .utils import create_files, verify_compressed


@ddt.ddt
class TestUtilities(unittest.TestCase):
    """Utilities for Enterprise reporting scripts."""

    SCARY_DICTIONARY = {
        'key1': {
            'key1-1': [
                'key1-1-v0',
                'key1-1-v1',
                'key1-1-v2',
            ],
            'key1-2': 'key1-2-v',
            'key1-3': {
                'key1-3-1': 'key1-3-1-v',
                'key1-3-2': [
                    'key1-3-2-v0',
                    'key1-3-2-v1',
                ]
            }
        },
        'key2': [
            'key2-v0',
            'key2-v1',
        ],
        'key3': True,
        'key4': 'key4-v',
        'key5': 'key5-v',
        'key6': [
            {
                'key6-1-1': 'key6-1-1-v',
                'key6-1-2': 'key6-1-2-v',
                'key6-1-3': [
                    'key6-1-3-v0',
                    'key6-1-3-v1',
                    'key6-1-3-v2',
                ],
            },
            {
                'key6-2-1': 'key6-2-1-v',
                'key6-2-2': 'key6-2-2-v',
                'key6-2-3': [
                    'key6-2-3-v0',
                    'key6-2-3-v1',
                    'key6-2-3-v2',
                ],
            },
        ]
    }

    @ddt.data(
        (
            'key',
            SCARY_DICTIONARY,
            [
                'key1_key1-1_0',
                'key1_key1-1_1',
                'key1_key1-1_2',
                'key1_key1-2',
                'key1_key1-3_key1-3-1',
                'key1_key1-3_key1-3-2_0',
                'key1_key1-3_key1-3-2_1',
                'key2_0',
                'key2_1',
                'key3',
                'key4',
                'key5',
                'key6_0_key6-1-1',
                'key6_0_key6-1-2',
                'key6_0_key6-1-3_0',
                'key6_0_key6-1-3_1',
                'key6_0_key6-1-3_2',
                'key6_1_key6-2-1',
                'key6_1_key6-2-2',
                'key6_1_key6-2-3_0',
                'key6_1_key6-2-3_1',
                'key6_1_key6-2-3_2',
            ],
        ),
        (
            'value',
            SCARY_DICTIONARY,
            [
                'key1-1-v0',
                'key1-1-v1',
                'key1-1-v2',
                'key1-2-v',
                'key1-3-1-v',
                'key1-3-2-v0',
                'key1-3-2-v1',
                'key2-v0',
                'key2-v1',
                True,
                'key4-v',
                'key5-v',
                'key6-1-1-v',
                'key6-1-2-v',
                'key6-1-3-v0',
                'key6-1-3-v1',
                'key6-1-3-v2',
                'key6-2-1-v',
                'key6-2-2-v',
                'key6-2-3-v0',
                'key6-2-3-v1',
                'key6-2-3-v2'
            ],
        ),
    )
    @ddt.unpack
    def test_flatten_dict(self, target, dictionary, flattened):
        """The dictionary, regardless of how complex, has its keys flattened."""
        assert utils.flatten_dict(dictionary, target=target) == flattened

    @ddt.data(
        {
            'H': [
                {'H': 'i'},
                'Hi',
                {'H': 'i'},
            ],
        },
        {
            'i': [
                ['Hi']
            ],
        },
        {
            'Hi': [
                ['H'],
                ['i'],
                'Hi',
            ]
        }
    )
    def test_flatten_dict_weird_cases(self, dictionary):
        """The dictionary, when weird, raises an error."""
        with self.assertRaises(NotImplementedError):
            utils.flatten_dict(dictionary)

    def test_is_current_time_in_schedule(self):
        """
        Verify that is_current_time_in_schedule works as expected for daily frequency.
        """
        est_timezone = pytz.timezone('US/Eastern')
        current_est_time = datetime.datetime.now(est_timezone)
        assert utils.is_current_time_in_schedule(
            current_est_time,
            utils.FREQUENCY_TYPE_DAILY,
            current_est_time.hour,
            current_est_time.day,
            current_est_time.weekday()
        )


@ddt.ddt
class TestCompressEncrypt(unittest.TestCase):
    """
    Tests `compress_and_encrypt` works correctly.
    """
    def pgpy_create_key(self, username):
        key = pgpy.PGPKey.new(PubKeyAlgorithm.RSAEncryptOrSign, 4096)
        uid = pgpy.PGPUID.new(username, comment='Unknown Person', email=username+'@unknown.com')
        key.add_uid(uid, usage={KeyFlags.Sign, KeyFlags.EncryptCommunications, KeyFlags.EncryptStorage},
                    hashes=[HashAlgorithm.SHA256, HashAlgorithm.SHA384, HashAlgorithm.SHA512, HashAlgorithm.SHA224],
                    ciphers=[SymmetricKeyAlgorithm.AES256, SymmetricKeyAlgorithm.AES192, SymmetricKeyAlgorithm.AES128],
                    compression=[CompressionAlgorithm.ZLIB, CompressionAlgorithm.BZ2, CompressionAlgorithm.ZIP,
                                 CompressionAlgorithm.Uncompressed])
        return key

    @ddt.data(
        [
            {
                'name': 'lord-of-the-rings.txt',
                'size': 1000
            },
        ],
        [
            {
                'name': 'lord-of-the-rings.txt',
                'size': 1000
            },
            {
                'name': 'harry-potter-and-deathly-hollows.txt',
                'size': 500
            },
        ],
    )
    def test_compress_and_encrypt(self, files_data):
        """
        Test that files are correctly compressed.
        """
        files, total_original_size = create_files(files_data)

        password = 'frodo-baggins'
        compressed_file = utils.compress_and_encrypt(
            [file['file'] for file in files],
            password
        )

        verify_compressed(self, compressed_file, files, total_original_size, password)

    @ddt.data(
        [
            {
                'name': 'lord-of-the-rings.txt',
                'size': 1000
            },
        ],
        [
            {
                'name': 'lord-of-the-rings.txt',
                'size': 1000
            },
            {
                'name': 'harry-potter-and-deathly-hollows.txt',
                'size': 500
            },
        ],
    )
    def test_encryption(self, files_data):
        """
        Test the successful decryption with a valid key
        and an unsuccessful decryption when an incorrect key.
        """
        files, size = create_files(files_data)
        password = 'low-complexity'
        correct_key = self.pgpy_create_key('JohnDoe')
        wrong_key = self.pgpy_create_key('JohnDoe2')

        encrypted_file_name = utils.compress_and_encrypt(
            [file['file'] for file in files],
            password,
            str(correct_key.pubkey)
        )
        message = pgpy.PGPMessage.from_file(encrypted_file_name)
        decrypted_message = correct_key.decrypt(message)
        self.assertIsInstance(decrypted_message, pgpy.PGPMessage)

        with self.assertRaises(PGPError):
            wrong_key.decrypt(message)

    @ddt.data(
        '_catalog_json_2019-09-30.json',
        '_catalog_json.json'
    )
    def test_compression_file_name(self, file_suffix):
        """
        Tests that files are compressed with the correct file name, even when date is included or excluded
        """
        tf = tempfile.NamedTemporaryFile(suffix=file_suffix)
        tf.write(b'randomtext54321')
        actual_file_name = utils._get_compressed_file([tf])
        expected_file_name = tf.name.split('.json', 1)[0] + '.zip'
        assert actual_file_name == expected_file_name


class TestPrepareAttachments(unittest.TestCase):

    def test_prepare_attachments(self):
        """
        prepare_attachments should properly create MIMEApplication
        depending on whether or not attachment data is provided
        """
        file_on_filesystem = tempfile.NamedTemporaryFile()
        file_on_filesystem_name = file_on_filesystem.name

        attachment_data = OrderedDict([
            (file_on_filesystem_name, None),
            ('my_test.csv', 'some,csv,data\n'),
        ])
        attachments = utils.prepare_attachments(attachment_data)

        for index, filename in enumerate([file_on_filesystem_name, 'my_test.csv']):
            expected_header = 'attachment; filename="{}"'.format(
                os.path.basename(filename)
            )
            assert attachments[index].get('Content-Disposition') == expected_header


class TestRetryOnException(unittest.TestCase):
    """
    Test that the decorator `retry_on_exception` works correctly.
    """

    def test_retry_on_exception(self):
        """
        retry_on_exception should retry the function the given
        number of times if an exception is raised.
        """

        @utils.retry_on_exception(max_retries=1, delay=1, backoff=1)
        def raise_exception():
            raise Exception('test')

        # The function should be retried once after an exception,
        # then raise the exception.
        with self.assertRaises(Exception):
            raise_exception()



# Generated by Django 3.2.20 on 2023-10-19 15:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise_data', '0037_alter_enterpriseenrollment_consent_granted'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterpriseoffer',
            name='export_timestamp',
            field=models.DateTimeField(null=True),
        ),
    ]



"""
Caching related utility classes and functions.
"""
import hashlib

from edx_django_utils.cache import TieredCache

DEFAULT_TIMEOUT = 60 * 60  # 1 hour


def get_key(*args, **kwargs):
    """
    Get MD5 encoded cache key for given positional and keyword arguments.

    MD5 encrytion is applied to a key that is generated by concatenating the positional and keyword arguments.
    Following is the format of the generated key from arguments before applying the MD5 encryption.
        arg1__arg2__key1:value1__key2:value2 ...

    Example:
        >>> get_key('ecommerce', site_domain='example.com', resource='catalogs')
        1892cd85a30b8fc9180369c17b472c38
        >>> # The generated key for the above call before applying MD5 encryption will be as follows
        >>> # "ecommerce__site_domain:example.com__resource:catalogs"

    Arguments:
        *args: Arguments that need to be present in cache key.
        **kwargs: Key word arguments that need to be present in cache key.

    Returns:
         (str): An MD5 encoded key uniquely identified by the key word arguments.
    """
    key = '{}__{}'.format(
        '__'.join(map(str, args)),
        '__'.join(['{}:{}'.format(item, str(value)) for item, value in kwargs.items()])
    )

    return hashlib.md5(key.encode('utf-8')).hexdigest()


def get(key):
    """
    Get value from cache for given key.

    Returns:
        (CachedResponse): CachedResponse object.
    """
    return TieredCache.get_cached_response(key)


def set(key, value, timeout=DEFAULT_TIMEOUT):  # pylint: disable=redefined-builtin
    """
    Set value in cache for given key.

    Arguments:
        key (str): Cache key.
        value (object): Value to be stored in cache.
        timeout (int): Cache timeout in seconds.
    """
    TieredCache.set_all_tiers(key, value, django_cache_timeout=timeout)



"""
Management command for creating enterprise learner enrollments
"""


import sys

from django.core.management.base import BaseCommand, CommandError

import enterprise_data.tests.test_utils


class Command(BaseCommand):
    """ Management command for creating dummy `EnterpriseLearnerEnrollment` instances for (LPR) V1."""

    help = 'Creates an EnterpriseLearnerEnrollment with randomized attributes'

    def add_arguments(self, parser):
        parser.add_argument(
            'enterprise_customer_uuid',
            type=str,
            help='UUID for an enterprise'
        )
        parser.add_argument(
            'enterprise_user_id',
            type=int,
            default=None,
            help='enterprise_user_id for an enterprise_user'
        )
        # this is optional and will be marked as False if not available in argumnets.
        parser.add_argument('--consent_granted', action='store_true')

    def handle(self, *args, **options):
        enterprise_customer_uuid = options['enterprise_customer_uuid']
        enterprise_user_id = options.get('enterprise_user_id')
        is_consent_granted = options.get('consent_granted')

        try:
            enterprise_data.tests.test_utils.EnterpriseLearnerEnrollmentFactory(
                enterprise_customer_uuid=enterprise_customer_uuid,
                enterprise_user_id=enterprise_user_id,
                is_consent_granted=is_consent_granted,
            )
            info = (
                '\nCreated EnterpriseLearnerEnrollment with enterprise_customer_uuid '
                '{} for EnterpriseUser with enterprise_user_id of {} and consent {}\n\n'.format(
                    enterprise_customer_uuid,
                    enterprise_user_id,
                    is_consent_granted,
                )
            )
            sys.stdout.write(info)
        except Exception as exc:
            info = (
                'Error trying to create EnterpriseUser with uuid '
                '{}: {}'.format(enterprise_customer_uuid, exc)
            )
            raise CommandError(info) from exc



"""
Module containing queries for the skills_daily_rollup_admin_dash table.
"""


class SkillsDailyRollupAdminDashQueries:
    """
    Queries related to the skills_daily_rollup_admin_dash table.
    """
    @staticmethod
    def get_top_skills():
        """
        Get the query to fetch the top skills for an enterprise customer.
        """
        return """
            SELECT
                skill_name,
                skill_type,
                SUM(enrolls) AS enrolls,
                SUM(completions) AS completions
            FROM
                skills_daily_rollup_admin_dash
            WHERE
                enterprise_customer_uuid=%(enterprise_customer_uuid)s AND
                date BETWEEN %(start_date)s AND %(end_date)s
            GROUP BY
                skill_name, skill_type
            ORDER BY
                enrolls DESC, completions DESC;
        """

    @staticmethod
    def get_top_skills_by_enrollment():
        """
        Get the query to fetch the top skills by enrollment for an enterprise customer.
        """
        return """
            WITH TopSkills AS (
                -- Get top 10 skills by total enrollments
                SELECT
                    skill_name,
                    SUM(enrolls) AS total_enrollment_count
                FROM
                    skills_daily_rollup_admin_dash
                WHERE
                    enterprise_customer_uuid=%(enterprise_customer_uuid)s
                    AND date BETWEEN %(start_date)s AND %(end_date)s
                GROUP BY
                    skill_name
                ORDER BY
                    total_enrollment_count DESC
                LIMIT 10
            )
            SELECT
                sd.skill_name,
                CASE
                    WHEN sd.primary_subject_name IN (
                        'business-management', 'computer-science',
                        'data-analysis-statistics', 'engineering', 'communication'
                    ) THEN sd.primary_subject_name
                    ELSE 'other'
                END AS subject_name,
                SUM(sd.enrolls) AS count
            FROM
                skills_daily_rollup_admin_dash AS sd
            JOIN
                TopSkills AS ts ON sd.skill_name = ts.skill_name
            WHERE
                sd.enterprise_customer_uuid=%(enterprise_customer_uuid)s
                AND date BETWEEN %(start_date)s AND %(end_date)s
            GROUP BY
                sd.skill_name, subject_name
            ORDER BY
                total_enrollment_count DESC;
        """

    @staticmethod
    def get_top_skills_by_completion():
        """
        Get the query to fetch the top skills by completion for an enterprise customer.
        """
        return """
            WITH TopSkills AS (
                -- Get top 10 skills by total completions
                SELECT
                    skill_name,
                    SUM(completions) AS total_completion_count
                FROM
                    skills_daily_rollup_admin_dash
                WHERE
                    enterprise_customer_uuid=%(enterprise_customer_uuid)s
                    AND date BETWEEN %(start_date)s AND %(end_date)s
                GROUP BY
                    skill_name
                ORDER BY
                    total_completion_count DESC
                LIMIT 10
            )
            SELECT
                sd.skill_name,
                CASE
                    WHEN sd.primary_subject_name IN (
                        'business-management', 'computer-science',
                        'data-analysis-statistics', 'engineering', 'communication'
                    ) THEN sd.primary_subject_name
                    ELSE 'other'
                END AS subject_name,
                SUM(sd.completions) AS count
            FROM
                skills_daily_rollup_admin_dash AS sd
            JOIN
                TopSkills AS ts ON sd.skill_name = ts.skill_name
            WHERE
                sd.enterprise_customer_uuid=%(enterprise_customer_uuid)s
                AND date BETWEEN %(start_date)s AND %(end_date)s
            GROUP BY
                sd.skill_name, subject_name
            ORDER BY
                total_completion_count DESC;
        """



"""
Enterprise Data Paginators.
"""


from edx_rest_framework_extensions.paginators import DefaultPagination


class EnterpriseEnrollmentsPagination(DefaultPagination):
    """
    Pagination format used by Enterprise Enrollments API.
    """

    max_page_size = 1000



# Generated by Django 1.11.5 on 2018-05-24 22:04


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise_data', '0004_auto_20180508_1652'),
    ]

    operations = [
        migrations.AlterField(
            model_name='enterpriseenrollment',
            name='enterprise_id',
            field=models.UUIDField(),
        ),
    ]



"""
Mock data for admin analytics tests.
"""

ENROLLMENTS = [
    {
        "enterprise_customer_name": "Hill Ltd",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "lms_enrollment_id": 1013,
        "user_id": 8907,
        "email": "rebeccanelson@example.com",
        "course_key": "hEmW+tvk03",
        "courserun_key": "course-v1:hEmW+tvk03+1T9889",
        "course_id": "1681",
        "course_subject": "business-management",
        "course_title": "Re-engineered tangible approach",
        "enterprise_enrollment_date": "2021-07-04",
        "lms_enrollment_mode": "verified",
        "enroll_type": "certificate",
        "program_title": "Non-Program",
        "date_certificate_awarded": "2021-08-25",
        "grade_percent": 0.99,
        "cert_awarded": 1,
        "date_certificate_created_raw": "2021-08-25",
        "passed_date_raw": "2021-08-25",
        "passed_date": "2021-08-25",
        "has_passed": 1,
    },
    {
        "enterprise_customer_name": "Hill Ltd",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "lms_enrollment_id": 9172,
        "user_id": 8369,
        "email": "taylorjames@example.com",
        "course_key": "hEmW+tvk03",
        "courserun_key": "course-v1:hEmW+tvk03+1T9889",
        "course_id": "1681",
        "course_subject": "business-management",
        "course_title": "Re-engineered tangible approach",
        "enterprise_enrollment_date": "2021-07-03",
        "lms_enrollment_mode": "verified",
        "enroll_type": "certificate",
        "program_title": "Non-Program",
        "date_certificate_awarded": "2021-09-01",
        "grade_percent": 0.93,
        "cert_awarded": 1,
        "date_certificate_created_raw": "2021-09-01",
        "passed_date_raw": "2021-09-01",
        "passed_date": "2021-09-01",
        "has_passed": 1,
    },
    {
        "enterprise_customer_name": "Hill Ltd",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "lms_enrollment_id": 9552,
        "user_id": 8719,
        "email": "ssmith@example.com",
        "course_key": "qZJC+KFX86",
        "courserun_key": "course-v1:qZJC+KFX86+1T8918",
        "course_id": "1725",
        "course_subject": "medicine",
        "course_title": "Secured static capability",
        "enterprise_enrollment_date": "2021-05-11",
        "lms_enrollment_mode": "verified",
        "enroll_type": "certificate",
        "program_title": "Non-Program",
        "date_certificate_awarded": None,
        "grade_percent": 0.0,
        "cert_awarded": 0,
        "date_certificate_created_raw": None,
        "passed_date_raw": None,
        "passed_date": None,
        "has_passed": 0,
    },
    {
        "enterprise_customer_name": "Hill Ltd",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "lms_enrollment_id": 3436,
        "user_id": 3125,
        "email": "kathleenmartin@example.com",
        "course_key": "QWXx+Jqz64",
        "courserun_key": "course-v1:QWXx+Jqz64+1T9449",
        "course_id": "4878",
        "course_subject": "social-sciences",
        "course_title": "Horizontal solution-oriented hub",
        "enterprise_enrollment_date": "2020-04-03",
        "lms_enrollment_mode": "verified",
        "enroll_type": "certificate",
        "program_title": "Non-Program",
        "date_certificate_awarded": None,
        "grade_percent": 0.0,
        "cert_awarded": 0,
        "date_certificate_created_raw": None,
        "passed_date_raw": None,
        "passed_date": None,
        "has_passed": 0,
    },
    {
        "enterprise_customer_name": "Hill Ltd",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "lms_enrollment_id": 5934,
        "user_id": 4853,
        "email": "amber79@example.com",
        "course_key": "NOGk+UVD31",
        "courserun_key": "course-v1:NOGk+UVD31+1T4956",
        "course_id": "4141",
        "course_subject": "communication",
        "course_title": "Streamlined zero-defect attitude",
        "enterprise_enrollment_date": "2020-04-08",
        "lms_enrollment_mode": "verified",
        "enroll_type": "certificate",
        "program_title": "Non-Program",
        "date_certificate_awarded": None,
        "grade_percent": 0.0,
        "cert_awarded": 0,
        "date_certificate_created_raw": None,
        "passed_date_raw": None,
        "passed_date": None,
        "has_passed": 0,
    },
]

ENGAGEMENTS = [
    {
        "user_id": 2340,
        "email": "padillamichelle@example.org",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-08-05",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 1,
        "is_engaged_video": 1,
        "is_engaged_forum": 0,
        "is_engaged_problem": 1,
        "is_active": 1,
        "learning_time_hours": 3724/3600,
        "learning_time_seconds": 3724,
    },
    {
        "user_id": 1869,
        "email": "yallison@example.org",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-07-27",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 1,
        "is_engaged_video": 1,
        "is_engaged_forum": 0,
        "is_engaged_problem": 1,
        "is_active": 1,
        "learning_time_hours": 4335/3600,
        "learning_time_seconds": 4335,
    },
    {
        "user_id": 6962,
        "email": "weaverpatricia@example.net",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-08-05",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 1,
        "is_engaged_video": 1,
        "is_engaged_forum": 0,
        "is_engaged_problem": 1,
        "is_active": 1,
        "learning_time_hours": 9441/3600,
        "learning_time_seconds": 9441,
    },
    {
        "user_id": 6798,
        "email": "seth57@example.org",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-08-21",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 1,
        "is_engaged_video": 1,
        "is_engaged_forum": 1,
        "is_engaged_problem": 1,
        "is_active": 1,
        "learning_time_hours": 9898/3600,
        "learning_time_seconds": 9898,
    },
    {
        "user_id": 8530,
        "email": "mackwilliam@example.com",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-10-24",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 0,
        "is_engaged_video": 0,
        "is_engaged_forum": 0,
        "is_engaged_problem": 0,
        "is_active": 1,
        "learning_time_hours": 0,
        "learning_time_seconds": 0,
    },
    {
        "user_id": 7584,
        "email": "graceperez@example.com",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2022-05-17",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 0,
        "is_engaged_video": 0,
        "is_engaged_forum": 0,
        "is_engaged_problem": 0,
        "is_active": 1,
        "learning_time_hours": 21/3600,
        "learning_time_seconds": 21,
    },
    {
        "user_id": 68,
        "email": "yferguson@example.net",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-09-02",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 1,
        "is_engaged_video": 1,
        "is_engaged_forum": 0,
        "is_engaged_problem": 1,
        "is_active": 1,
        "learning_time_hours": 4747/3600,
        "learning_time_seconds": 4747,
    },
    {
        "user_id": 4278,
        "email": "caseyjohnny@example.com",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "Kcpr+XoR30",
        "enroll_type": "certificate",
        "activity_date": "2022-05-20",
        "course_title": "Assimilated even-keeled focus group",
        "course_subject": "engineering",
        "is_engaged": 0,
        "is_engaged_video": 0,
        "is_engaged_forum": 0,
        "is_engaged_problem": 0,
        "is_active": 1,
        "learning_time_hours": 0,
        "learning_time_seconds": 0,
    },
    {
        "user_id": 8726,
        "email": "webertodd@example.com",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-09-21",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 1,
        "is_engaged_video": 1,
        "is_engaged_forum": 1,
        "is_engaged_problem": 1,
        "is_active": 1,
        "learning_time_hours": 5285/3600,
        "learning_time_seconds": 5285,
    },
    {
        "user_id": 282,
        "email": "crystal86@example.net",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-07-11",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 0,
        "is_engaged_video": 0,
        "is_engaged_forum": 0,
        "is_engaged_problem": 0,
        "is_active": 1,
        "learning_time_hours": 0,
        "learning_time_seconds": 0,
    },
    {
        "user_id": 2731,
        "email": "paul77@example.org",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-07-26",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 1,
        "is_engaged_video": 1,
        "is_engaged_forum": 1,
        "is_engaged_problem": 1,
        "is_active": 1,
        "learning_time_hours": 15753/3600,
        "learning_time_seconds": 15753,
    },
    {
        "user_id": 4038,
        "email": "samanthaclarke@example.org",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-07-19",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 0,
        "is_engaged_video": 0,
        "is_engaged_forum": 0,
        "is_engaged_problem": 0,
        "is_active": 1,
        "learning_time_hours": 29/3600,
        "learning_time_seconds": 29,
    }
]


LEADERBOARD_RESPONSE = [
    {
        "email": "paul77@example.org",
        "sessions": 1,
        "learning_time_hours": 4.4,
        "average_session_length": 4.4,
        "course_completions": None,
    },
    {
        "email": "seth57@example.org",
        "sessions": 1,
        "learning_time_hours": 2.7,
        "average_session_length": 2.7,
        "course_completions": None,
    },
    {
        "email": "weaverpatricia@example.net",
        "sessions": 1,
        "learning_time_hours": 2.6,
        "average_session_length": 2.6,
        "course_completions": None,
    },
    {
        "email": "webertodd@example.com",
        "sessions": 1,
        "learning_time_hours": 1.5,
        "average_session_length": 1.5,
        "course_completions": None,
    },
    {
        "email": "yferguson@example.net",
        "sessions": 1,
        "learning_time_hours": 1.3,
        "average_session_length": 1.3,
        "course_completions": None,
    },
    {
        "email": "yallison@example.org",
        "sessions": 1,
        "learning_time_hours": 1.2,
        "average_session_length": 1.2,
        "course_completions": None,
    },
    {
        "email": "padillamichelle@example.org",
        "sessions": 1,
        "learning_time_hours": 1.0,
        "average_session_length": 1.0,
        "course_completions": None,
    },
    {
        "email": "caseyjohnny@example.com",
        "sessions": 0,
        "learning_time_hours": 0.0,
        "average_session_length": 0.0,
        "course_completions": None,
    },
    {
        "email": "crystal86@example.net",
        "sessions": 0,
        "learning_time_hours": 0.0,
        "average_session_length": 0.0,
        "course_completions": None,
    },
    {
        "email": "graceperez@example.com",
        "sessions": 0,
        "learning_time_hours": 0.0,
        "average_session_length": 0.0,
        "course_completions": None,
    },
    {
        "email": "mackwilliam@example.com",
        "sessions": 0,
        "learning_time_hours": 0.0,
        "average_session_length": 0.0,
        "course_completions": None,
    },
    {
        "email": "samanthaclarke@example.org",
        "sessions": 0,
        "learning_time_hours": 0.0,
        "average_session_length": 0.0,
        "course_completions": None,
    },
]


TOP_SKILLS = [
    {
        "skill_name": "Python (Programming Language)",
        "skill_type": "Specialized Skill",
        "enrolls": 19027.0,
        "completions": 3004.0
    },
    {
        "skill_name": "Data Science",
        "skill_type": "Specialized Skill",
        "enrolls": 13756.0,
        "completions": 1517.0
    },
    {
        "skill_name": "Algorithms",
        "skill_type": "Specialized Skill",
        "enrolls": 12776.0,
        "completions": 1640.0
    },
]
TOP_SKILLS_BY_ENROLLMENTS = [
    {
        "skill_name": "Python (Programming Language)",
        "subject_name": "business-management",
        "count": 313.0
    },
    {
        "skill_name": "Machine Learning",
        "subject_name": "business-management",
        "count": 442.0
    },
    {
        "skill_name": "Computer Science",
        "subject_name": "business-management",
        "count": 39.0
    },
]
TOP_SKILLS_BY_COMPLETIONS = [
    {
        "skill_name": "Python (Programming Language)",
        "subject_name": "business-management",
        "count": 21.0
    },
    {
        "skill_name": "SQL (Programming Language)",
        "subject_name": "business-management",
        "count": 11.0
    },
    {
        "skill_name": "Algorithms",
        "subject_name": "business-management",
        "count": 15.0
    },
]



"""
Tests for create_dummy_data_lpr_v1 management command
"""
import uuid
from unittest import TestCase

from pytest import mark

from django.core.management import call_command

from enterprise_data.models import EnterpriseLearner, EnterpriseLearnerEnrollment


@mark.django_db
class TestCreateEnterpriseLearnerCommand(TestCase):
    """
    Tests to validate the behavior of `./manage.py create_dummy_data_lpr_v1` management command.
    """

    def setUp(self):
        super().setUp()
        self.enterprise_customer_uuid = uuid.uuid4()

    def test_create_enterprise_learners_with_enrollments(self):
        """
        Management command should successfully be able to create 10 EnterpriseLearners with 5 enrollments each
        """
        assert EnterpriseLearner.objects.count() == 0
        assert EnterpriseLearnerEnrollment.objects.count() == 0

        args = [self.enterprise_customer_uuid]
        call_command('create_dummy_data_lpr_v1', *args)

        assert EnterpriseLearner.objects.count() == 10
        assert EnterpriseLearnerEnrollment.objects.count() == 50



from django.db import migrations

from enterprise_data_roles.constants import ROLE_BASED_ACCESS_CONTROL_SWITCH


def create_switch(apps, schema_editor):
    """Create the `role_based_access_control` switch if it does not already exist."""
    Switch = apps.get_model('waffle', 'Switch')
    Switch.objects.update_or_create(name=ROLE_BASED_ACCESS_CONTROL_SWITCH, defaults={'active': False})


def delete_switch(apps, schema_editor):
    """Delete the `role_based_access_control` switch."""
    Switch = apps.get_model('waffle', 'Switch')
    Switch.objects.filter(name=ROLE_BASED_ACCESS_CONTROL_SWITCH).delete()


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise_data', '0022_remove_role_based_access_control_switch'),
        ('enterprise_data_roles', '0002_add_enterprise_data_feature_roles'),
        ('waffle', '0001_initial')
    ]

    operations = [
        migrations.RunPython(create_switch, delete_switch),
    ]



"""
Module for interacting with the skills_daily_rollup_admin_dash table.
"""
from datetime import date
from uuid import UUID

from enterprise_data.admin_analytics.database.queries.skills_daily_rollup_admin_dash import (
    SkillsDailyRollupAdminDashQueries,
)
from enterprise_data.admin_analytics.database.tables.base import BaseTable
from enterprise_data.admin_analytics.database.utils import run_query
from enterprise_data.cache.decorators import cache_it


class SkillsDailyRollupAdminDashTable(BaseTable):
    """
    Class for communicating with the skills_daily_rollup_admin_dash table.
    """
    queries = SkillsDailyRollupAdminDashQueries()

    @cache_it()
    def get_top_skills(self, enterprise_customer_uuid: UUID, start_date: date, end_date: date):
        """
        Get the top skills for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            list<dict>: A list of dictionaries containing the skill_name, skill_type, enrolls and completions.
        """
        return run_query(
            query=self.queries.get_top_skills(),
            params={
                'enterprise_customer_uuid': enterprise_customer_uuid,
                'start_date': start_date,
                'end_date': end_date,
            },
            as_dict=True,
        )

    @cache_it()
    def get_top_skills_by_enrollment(
        self,
        enterprise_customer_uuid: UUID,
        start_date: date,
        end_date: date
    ):
        """
        Get the top skills by enrollments for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            list<dict>: A list of dictionaries containing the skill_name, subject_name, count.
        """
        return run_query(
            query=self.queries.get_top_skills_by_enrollment(),
            params={
                'enterprise_customer_uuid': enterprise_customer_uuid,
                'start_date': start_date,
                'end_date': end_date,
            },
            as_dict=True,
        )

    @cache_it()
    def get_top_skills_by_completion(
        self,
        enterprise_customer_uuid: UUID,
        start_date: date,
        end_date: date
    ):
        """
        Get the top skills by completion for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            list<dict>: A list of dictionaries containing the skill_name, subject_name, count.
        """
        return run_query(
            query=self.queries.get_top_skills_by_completion(),
            params={
                'enterprise_customer_uuid': enterprise_customer_uuid,
                'start_date': start_date,
                'end_date': end_date,
            },
            as_dict=True,
        )



"""
Decorators for caching the result of a function.
"""
from functools import wraps
from logging import getLogger

from enterprise_data import cache

LOGGER = getLogger(__name__)


def cache_it(timeout=cache.DEFAULT_TIMEOUT):
    """
    Function to return the decorator to cache the result of a method.

    Note: This decorator will only work for class methods.

    Arguments:
        timeout (int): Cache timeout in seconds.

    Returns:
        (function): Decorator function.
    """

    def inner_decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            """
            Wrapper function to cache the result of the function.
            """
            cache_key = cache.get_key(func.__name__, *args, **kwargs)
            cached_response = cache.get(cache_key)
            if cached_response.is_found:
                LOGGER.info("[ANALYTICS]: Cache hit for key: (%s)", (func.__name__, args, kwargs))
                return cached_response.value

            LOGGER.info("[ANALYTICS]: Cache miss for key: (%s)", (func.__name__, args, kwargs))
            result = func(self, *args, **kwargs)
            cache.set(cache_key, result, timeout=timeout)
            return result
        return wrapper
    return inner_decorator



"""
Management command for pre-warming the analytics cache for large enterprises.
"""
from datetime import date

from django.core.management.base import BaseCommand, CommandError

from enterprise_data.admin_analytics.database.tables import (
    FactEngagementAdminDashTable,
    FactEnrollmentAdminDashTable,
    SkillsDailyRollupAdminDashTable,
)


class Command(BaseCommand):
    """
    Add cache entries for analytics related data for a large enterprise.

    The top enterprises will be the ones with the most enrollments.
    """
    help = 'Pre-warm the analytics cache for a large enterprises.'

    @staticmethod
    def __cache_enrollment_data(enterprise_customer_uuid):
        """
        Helper method to cache all the enrollment related data for the given enterprise.

        Arguments:
            enterprise_customer_uuid (str): The UUID of the enterprise customer.
        """
        enterprise_enrollment_table = FactEnrollmentAdminDashTable()
        start_date, _ = enterprise_enrollment_table.get_enrollment_date_range(
            enterprise_customer_uuid,
        )
        end_date = date.today()
        enterprise_enrollment_table.get_enrollment_count(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )
        page_size = 100
        enterprise_enrollment_table.get_all_enrollments(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
            limit=page_size,
            offset=0,
        )
        enterprise_enrollment_table.get_enrollment_and_course_count(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )
        enterprise_enrollment_table.get_completion_count(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )
        enterprise_enrollment_table.get_top_courses_by_enrollments(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )
        enterprise_enrollment_table.get_top_subjects_by_enrollments(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )
        enterprise_enrollment_table.get_enrolment_time_series_data(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )

    @staticmethod
    def __cache_completions_data(enterprise_customer_uuid):
        """
        Helper method to cache all the completions related data for the given enterprise.

        Arguments:
            enterprise_customer_uuid (str): The UUID of the enterprise customer.
        """
        enterprise_enrollment_table = FactEnrollmentAdminDashTable()
        start_date, _ = enterprise_enrollment_table.get_enrollment_date_range(
            enterprise_customer_uuid,
        )
        end_date = date.today()

        page_size = 100
        enterprise_enrollment_table.get_all_completions(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
            limit=page_size,
            offset=0,
        )
        enterprise_enrollment_table.get_completion_count(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )
        enterprise_enrollment_table.get_top_courses_by_completions(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )
        enterprise_enrollment_table.get_top_subjects_by_completions(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )
        enterprise_enrollment_table.get_completions_time_series_data(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )

    @staticmethod
    def __cache_engagement_data(enterprise_customer_uuid):
        """
        Helper method to cache all the engagement related data for the given enterprise.

        Arguments:
            enterprise_customer_uuid (str): The UUID of the enterprise customer.
        """
        start_date, _ = FactEnrollmentAdminDashTable().get_enrollment_date_range(
            enterprise_customer_uuid,
        )
        end_date = date.today()
        enterprise_engagement_table = FactEngagementAdminDashTable()
        enterprise_engagement_table.get_learning_hours_and_daily_sessions(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )
        enterprise_engagement_table.get_engagement_count(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )
        page_size = 100
        enterprise_engagement_table.get_all_engagements(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
            limit=page_size,
            offset=0,
        )
        enterprise_engagement_table.get_top_courses_by_engagement(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )
        enterprise_engagement_table.get_top_subjects_by_engagement(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )
        enterprise_engagement_table.get_engagement_time_series_data(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )
        total_count = enterprise_engagement_table.get_leaderboard_data_count(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )
        enterprise_engagement_table.get_all_leaderboard_data(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
            limit=page_size,
            offset=0,
            total_count=total_count,
        )

    @staticmethod
    def __cache_skills_data(enterprise_customer_uuid):
        """
        Helper method to cache all the skills related data for the given enterprise.

        Arguments:
            enterprise_customer_uuid (str): The UUID of the enterprise customer.
        """
        start_date, _ = FactEnrollmentAdminDashTable().get_enrollment_date_range(
            enterprise_customer_uuid,
        )
        end_date = date.today()
        skills_table = SkillsDailyRollupAdminDashTable()
        skills_table.get_top_skills(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )
        skills_table.get_top_skills_by_enrollment(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )
        skills_table.get_top_skills_by_completion(
            enterprise_customer_uuid=enterprise_customer_uuid,
            start_date=start_date,
            end_date=end_date,
        )

    def handle(self, *args, **options):
        for enterprise_customer_uuid in FactEnrollmentAdminDashTable().get_top_enterprises():
            try:
                self.__cache_enrollment_data(enterprise_customer_uuid)
                self.__cache_completions_data(enterprise_customer_uuid)
                self.__cache_engagement_data(enterprise_customer_uuid)
                self.__cache_skills_data(enterprise_customer_uuid)
            except Exception as exc:
                info = (
                    'Error trying to add cache entries for enterprise '
                    '{}: {}'.format(enterprise_customer_uuid, exc)
                )
                raise CommandError(info) from exc



"""
Module containing queries for the fact_enrollment_admin_dash table.
"""
from .fact_engagement_admin_dash import FactEngagementAdminDashQueries
from .fact_enrollment_admin_dash import FactEnrollmentAdminDashQueries



"""
Tests for the `edx-enterprise` api v1 serializer module.
"""

import uuid

import ddt
from pytest import mark
from rest_framework.test import APITransactionTestCase

from enterprise_data.api.v1.serializers import EnterpriseLearnerEnrollmentSerializer, EnterpriseOfferSerializer
from enterprise_data.tests.test_utils import (
    EnterpriseLearnerEnrollmentFactory,
    EnterpriseLearnerFactory,
    EnterpriseOfferFactory,
)


@mark.django_db
@ddt.ddt
class TestEnterpriseLearnerEnrollmentSerializer(APITransactionTestCase):
    """
    Tests for `enterprise_learner_enrollment` API serializer.
    """

    def setUp(self):
        super().setUp()
        ent_user = EnterpriseLearnerFactory()
        self.enrollment = EnterpriseLearnerEnrollmentFactory(
            enterprise_user_id=ent_user.enterprise_user_id,
            is_consent_granted=True
        )

    @ddt.data(
        (0.0, 0.0),
        (3600, 1.0),
    )
    @ddt.unpack
    def test_total_learning_time_hours(self, total_learning_time_seconds, expected_total_learning_time_seconds):
        self.enrollment.total_learning_time_seconds = total_learning_time_seconds
        self.enrollment.save()
        serializer = EnterpriseLearnerEnrollmentSerializer(self.enrollment)
        assert serializer.data['total_learning_time_hours'] == expected_total_learning_time_seconds


@ddt.ddt
class TestEnterpriseOfferSerializer(APITransactionTestCase):
    """
    Tests for `enterprise_offer` API serializer.
    """

    @ddt.data(
        ('787f0fa7-f89a-4267-9e2b-fcd299e83918', '787f0fa7f89a42679e2bfcd299e83918'),
        ('56a779a5-21ec-4711-830f-6654d3faaa5f', '56a779a521ec4711830f6654d3faaa5f'),
        ('12345', '12345'),
    )
    @ddt.unpack
    def test_deserialize_offer_id_valid(self, offer_id_request, offer_id_db):
        data = {
            'offer_id': offer_id_request,
            'enterprise_customer_uuid': str(uuid.uuid4()),
        }
        serialized_offer = EnterpriseOfferSerializer(data=data)
        assert serialized_offer.is_valid()
        assert serialized_offer.validated_data['offer_id'] == offer_id_db

    @ddt.data(
        # Representations that could be real UUIDs should always be normalized to un-hyphenated UUIDs.
        ('787f0fa7f89a42679e2bfcd299e83918', '787f0fa7-f89a-4267-9e2b-fcd299e83918'),
        ('f0d5fcdbed8e4d408f3b79d1671f967f', 'f0d5fcdb-ed8e-4d40-8f3b-79d1671f967f'),
        ('c1d39f8a3fd246758bf7f56dd076694d', 'c1d39f8a-3fd2-4675-8bf7-f56dd076694d'),
        ('12345', '12345'),
    )
    @ddt.unpack
    def test_serialize_offer_id_valid(self, offer_id_db, offer_id_serialized):
        offer = EnterpriseOfferFactory(offer_id=offer_id_db)
        serialized_offer = EnterpriseOfferSerializer(offer)
        assert serialized_offer.data['offer_id'] == offer_id_serialized

    @ddt.data(
        # The following should fail validation because we want to avoid storing these into the DB.
        None,  # null values not allowed.
        '',  # empty values not allowed.
        'abc',  # Not exactly 36 characters but also obviously not an integer.
        '787f0fa7f89a42679e2bfcd299e83918',  # Not exactly 36 characters but also obviously not an integer.
        '11111111111',  # greater than 10 digit integer
        '111111111111111111111111111111111111',  # 36 digit integer
        '11111111111111111111111111111111',  # 32 digit integer
    )
    def test_deserialize_offer_id_invalid(self, offer_id_request):
        data = {
            'offer_id': offer_id_request,
            'enterprise_customer_uuid': str(uuid.uuid4()),
        }
        serializer = EnterpriseOfferSerializer(data=data)
        assert not serializer.is_valid()



# Generated by Django 4.2.16 on 2025-01-08 07:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise_data', '0045_alter_enterpriseexecedlcmoduleperformance_options_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='EnterpriseGroupMembership',
            fields=[
                ('group_membership_unique_id', models.CharField(max_length=64, primary_key=True, serialize=False)),
                ('is_applies_to_all_contexts', models.BooleanField(default=False)),
                ('enterprise_customer_id', models.UUIDField(db_index=True, null=True)),
                ('enterprise_group_name', models.CharField(max_length=255, null=True)),
                ('enterprise_group_uuid', models.UUIDField(null=True)),
                ('group_is_removed', models.BooleanField(default=False)),
                ('group_type', models.CharField(max_length=128, null=True)),
                ('activated_at', models.DateTimeField(null=True)),
                ('enterprise_customer_user_id', models.PositiveIntegerField(null=True)),
                ('membership_is_removed', models.BooleanField(default=False)),
                ('membership_status', models.CharField(max_length=128, null=True)),
                ('enterprise_group_membership_uuid', models.UUIDField(null=True)),
            ],
            options={
                'verbose_name': 'Group Membership',
                'verbose_name_plural': 'Group Memberships',
                'db_table': 'group_membership',
                'indexes': [models.Index(fields=['enterprise_group_uuid', 'group_type', 'enterprise_customer_user_id'], name='group_membe_enterpr_796f48_idx')],
            },
        ),
    ]



"""
Tests for create_enterprise_enrollment management command
"""
from unittest import TestCase, mock

from pytest import mark

from django.core.management import call_command

from enterprise_data.models import EnterpriseEnrollment, EnterpriseUser
from enterprise_data.tests.test_utils import EnterpriseUserFactory


@mark.django_db
class TestCreateEnterpriseEnrollmentCommand(TestCase):
    """ test class here """

    def setUp(self):
        super().setUp()
        self.uuid = 'a'*32
        self.enterprise_user = EnterpriseUserFactory(enterprise_id=self.uuid)

    def test_create_enterpriseenrollment(self):
        """
        Manangement command should successfully be able to create EnterpriseEnrollment
        """
        assert EnterpriseUser.objects.count() == 1
        assert EnterpriseEnrollment.objects.count() == 0

        args = [self.uuid, self.enterprise_user.enterprise_user_id]
        call_command('create_enterprise_enrollment', *args)

        assert EnterpriseEnrollment.objects.count() == 1
        assert EnterpriseEnrollment.objects.filter(
            enterprise_id=args[0]
        ).count() == 1

    def test_create_enterpriseenrollment_error(self):
        """
        Management command will not create enrollment if an error is thrown
        """
        assert EnterpriseEnrollment.objects.count() == 0

        args = [self.uuid, self.enterprise_user.enterprise_user_id]
        with mock.patch('enterprise_data.tests.test_utils.EnterpriseEnrollmentFactory') as mock_factory:
            mock_factory.side_effect = [Exception]
            with self.assertRaises(Exception):
                call_command('create_enterprise_enrollment', *args)

        assert EnterpriseEnrollment.objects.count() == 0



"""Unittests for analytics_enrollments.py"""
from datetime import datetime

import ddt
from mock import patch
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITransactionTestCase

from enterprise_data.admin_analytics.constants import ResponseType
from enterprise_data.tests.admin_analytics.mock_analytics_data import ENROLLMENTS
from enterprise_data.tests.mixins import JWTTestMixin
from enterprise_data.tests.test_utils import UserFactory
from enterprise_data_roles.constants import ENTERPRISE_DATA_ADMIN_ROLE
from enterprise_data_roles.models import EnterpriseDataFeatureRole, EnterpriseDataRoleAssignment


@ddt.ddt
class TestIndividualEnrollmentsAPI(JWTTestMixin, APITransactionTestCase):
    """
    Tests for AdvanceAnalyticsIndividualEnrollmentsView.
    """

    def setUp(self):
        """
        Setup method.
        """
        super().setUp()
        self.user = UserFactory(is_staff=True)
        role, __ = EnterpriseDataFeatureRole.objects.get_or_create(
            name=ENTERPRISE_DATA_ADMIN_ROLE
        )
        self.role_assignment = EnterpriseDataRoleAssignment.objects.create(
            role=role, user=self.user
        )
        self.client.force_authenticate(user=self.user)

        self.enterprise_uuid = "ee5e6b3a-069a-4947-bb8d-d2dbc323396c"
        self.set_jwt_cookie()

        self.url = reverse(
            "v1:enterprise-admin-analytics-enrollments",
            kwargs={"enterprise_uuid": self.enterprise_uuid},
        )

        get_enrollment_date_range_patcher = patch(
            'enterprise_data.api.v1.views.analytics_enrollments.FactEnrollmentAdminDashTable.get_enrollment_date_range',
            return_value=(datetime.now(), datetime.now())
        )

        get_enrollment_date_range_patcher.start()
        self.addCleanup(get_enrollment_date_range_patcher.stop)

    @patch('enterprise_data.api.v1.views.analytics_enrollments.FactEnrollmentAdminDashTable.get_enrollment_count')
    @patch('enterprise_data.api.v1.views.analytics_enrollments.FactEnrollmentAdminDashTable.get_all_enrollments')
    def test_get(self, mock_get_all_enrollments, mock_get_enrollment_count):
        """
        Test the GET method for the AdvanceAnalyticsIndividualEnrollmentsView works.
        """
        mock_get_all_enrollments.return_value = ENROLLMENTS
        mock_get_enrollment_count.return_value = len(ENROLLMENTS)

        response = self.client.get(self.url + '?page_size=2')
        assert response.status_code == status.HTTP_200_OK
        assert response["Content-Type"] == "application/json"
        data = response.json()
        assert data["next"] == f"http://testserver{self.url}?page=2&page_size=2"
        assert data["previous"] is None
        assert data["current_page"] == 1
        assert data["num_pages"] == 3
        assert data["count"] == 5

        response = self.client.get(self.url + '?page_size=2&page=2')
        assert response.status_code == status.HTTP_200_OK
        assert response["Content-Type"] == "application/json"
        data = response.json()
        assert data["next"] == f"http://testserver{self.url}?page=3&page_size=2"
        assert data["previous"] == f"http://testserver{self.url}?page_size=2"
        assert data["current_page"] == 2
        assert data["num_pages"] == 3
        assert data["count"] == 5

        response = self.client.get(self.url + '?page_size=2&page=3')
        assert response.status_code == status.HTTP_200_OK
        assert response["Content-Type"] == "application/json"
        data = response.json()
        assert data["next"] is None
        assert data["previous"] == f"http://testserver{self.url}?page=2&page_size=2"
        assert data["current_page"] == 3
        assert data["num_pages"] == 3
        assert data["count"] == 5

        response = self.client.get(self.url + '?page_size=5')
        assert response.status_code == status.HTTP_200_OK
        assert response["Content-Type"] == "application/json"
        data = response.json()
        assert data["next"] is None
        assert data["previous"] is None
        assert data["current_page"] == 1
        assert data["num_pages"] == 1
        assert data["count"] == 5

    @patch('enterprise_data.api.v1.views.analytics_enrollments.FactEnrollmentAdminDashTable.get_enrollment_count')
    @patch('enterprise_data.api.v1.views.analytics_enrollments.FactEnrollmentAdminDashTable.get_all_enrollments')
    def test_get_csv(self, mock_get_all_enrollments, mock_get_enrollment_count):
        """
        Test the GET method for the AdvanceAnalyticsIndividualEnrollmentsView return correct CSV data.
        """
        mock_get_all_enrollments.return_value = ENROLLMENTS
        mock_get_enrollment_count.return_value = len(ENROLLMENTS)
        response = self.client.get(self.url, {"response_type": ResponseType.CSV.value})
        assert response.status_code == status.HTTP_200_OK

        # verify the response headers
        assert response["Content-Type"] == "text/csv"

        # verify the response content
        content = b"".join(response.streaming_content).decode().splitlines()
        assert len(content) == 6

        # Verify CSV header.
        assert 'email,course_title,course_subject,enroll_type,enterprise_enrollment_date' == content[0]

        # verify the content
        assert (
            'rebeccanelson@example.com,Re-engineered tangible approach,business-management,certificate,2021-07-04'
            in content
        )
        assert (
            'taylorjames@example.com,Re-engineered tangible approach,business-management,certificate,2021-07-03'
            in content
        )
        assert (
            'ssmith@example.com,Secured static capability,medicine,certificate,2021-05-11'
            in content
        )
        assert (
            'amber79@example.com,Streamlined zero-defect attitude,communication,certificate,2020-04-08'
            in content
        )
        assert (
            'kathleenmartin@example.com,Horizontal solution-oriented hub,social-sciences,certificate,2020-04-03'
            in content
        )

    @ddt.data(
        {
            "params": {"start_date": 1},
            "error": {
                "start_date": [
                    "Date has wrong format. Use one of these formats instead: YYYY-MM-DD."
                ]
            },
        },
        {
            "params": {"end_date": 2},
            "error": {
                "end_date": [
                    "Date has wrong format. Use one of these formats instead: YYYY-MM-DD."
                ]
            },
        },
        {
            "params": {"start_date": "2024-01-01", "end_date": "2023-01-01"},
            "error": {
                "non_field_errors": [
                    "start_date should be less than or equal to end_date."
                ]
            },
        },
    )
    @ddt.unpack
    def test_get_invalid_query_params(self, params, error):
        """
        Test the GET method return correct error if any query param value is incorrect.
        """
        response = self.client.get(self.url, params)
        assert response.status_code == status.HTTP_400_BAD_REQUEST
        assert response.json() == error


@ddt.ddt
class TestEnrollmentStatsAPI(JWTTestMixin, APITransactionTestCase):
    """
    Tests for AdvanceAnalyticsEnrollmentStatsView.
    """

    def setUp(self):
        """
        Setup method.
        """
        super().setUp()
        self.user = UserFactory(is_staff=True)
        role, __ = EnterpriseDataFeatureRole.objects.get_or_create(
            name=ENTERPRISE_DATA_ADMIN_ROLE
        )
        self.role_assignment = EnterpriseDataRoleAssignment.objects.create(
            role=role, user=self.user
        )
        self.client.force_authenticate(user=self.user)

        self.enterprise_uuid = "ee5e6b3a-069a-4947-bb8d-d2dbc323396c"
        self.set_jwt_cookie()

        self.url = reverse(
            "v1:enterprise-admin-analytics-enrollments-stats",
            kwargs={"enterprise_uuid": self.enterprise_uuid},
        )

        get_enrollment_date_range_patcher = patch(
            'enterprise_data.api.v1.views.analytics_enrollments.FactEnrollmentAdminDashTable.get_enrollment_date_range',
            return_value=(datetime.now(), datetime.now())
        )

        get_enrollment_date_range_patcher.start()
        self.addCleanup(get_enrollment_date_range_patcher.stop)

    @patch(
        'enterprise_data.api.v1.views.analytics_enrollments.FactEnrollmentAdminDashTable.'
        'get_top_subjects_by_enrollments'
    )
    @patch(
        'enterprise_data.api.v1.views.analytics_enrollments.FactEnrollmentAdminDashTable.get_top_courses_by_enrollments'
    )
    @patch(
        'enterprise_data.api.v1.views.analytics_enrollments.FactEnrollmentAdminDashTable.get_enrolment_time_series_data'
    )
    def test_get(
            self,
            mock_get_enrolment_time_series_data,
            mock_get_top_courses_by_enrollments,
            mock_get_top_subjects_by_enrollments,
    ):
        """
        Test the GET method for the AdvanceAnalyticsEnrollmentStatsView works.
        """
        mock_get_enrolment_time_series_data.return_value = []
        mock_get_top_courses_by_enrollments.return_value = []
        mock_get_top_subjects_by_enrollments.return_value = []

        response = self.client.get(self.url)
        assert response.status_code == status.HTTP_200_OK
        assert response["Content-Type"] == "application/json"
        data = response.json()
        assert 'enrollments_over_time' in data
        assert 'top_courses_by_enrollments' in data
        assert 'top_subjects_by_enrollments' in data

    @ddt.data(
        {
            "params": {"start_date": 1},
            "error": {
                "start_date": [
                    "Date has wrong format. Use one of these formats instead: YYYY-MM-DD."
                ]
            },
        },
        {
            "params": {"end_date": 2},
            "error": {
                "end_date": [
                    "Date has wrong format. Use one of these formats instead: YYYY-MM-DD."
                ]
            },
        },
        {
            "params": {"start_date": "2024-01-01", "end_date": "2023-01-01"},
            "error": {
                "non_field_errors": [
                    "start_date should be less than or equal to end_date."
                ]
            },
        },
    )
    @ddt.unpack
    def test_get_invalid_query_params(self, params, error):
        """
        Test the GET method return correct error if any query param value is incorrect.
        """
        response = self.client.get(self.url, params)
        assert response.status_code == status.HTTP_400_BAD_REQUEST
        assert response.json() == error



"""
This module contains the database queries for the admin analytics.
"""
from .fact_engagement_admin_dash import FactEngagementAdminDashTable
from .fact_enrollment_admin_dash import FactEnrollmentAdminDashTable
from .skills_daily_rollup_admin_dash import SkillsDailyRollupAdminDashTable



"""
Module for interacting with the fact_enrollment_admin_dash table.
"""
from datetime import date, datetime
from uuid import UUID

from enterprise_data.cache.decorators import cache_it

from ..queries import FactEnrollmentAdminDashQueries
from ..utils import run_query
from .base import BaseTable


class FactEnrollmentAdminDashTable(BaseTable):
    """
    Class for communicating with the fact_enrollment_admin_dash table.
    """
    queries = FactEnrollmentAdminDashQueries()

    def get_top_enterprises(self, count=10):
        """
        Get the top enterprises by enrollments.

        Arguments:
            count (int): The number of enterprises to return.

        Returns:
            list<str>: A list of enterprise UUIDs.
        """
        result = run_query(
            query=self.queries.get_top_enterprises_query(count),
            as_dict=False,
        )
        return [row[0] for row in result]

    @cache_it()
    def get_enrollment_count(self, enterprise_customer_uuid: UUID, start_date: date, end_date: date):
        """
        Get the total number of enrollments for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            (int): The total number of enrollments.
        """
        results = run_query(
            query=self.queries.get_enrollment_count_query(),
            params={
                'enterprise_customer_uuid': enterprise_customer_uuid,
                'start_date': start_date,
                'end_date': end_date,
            }
        )
        if not results:
            return 0
        return int(results[0][0] or 0)

    @cache_it()
    def get_all_enrollments(
            self, enterprise_customer_uuid: UUID, start_date: date, end_date: date, limit: int, offset: int
    ):
        """
        Get all enrollments for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            limit (int): The maximum number of records to return.
            offset (int): The number of records to skip.

        Returns:
            list<dict>: A list of dictionaries containing the enrollment data.
        """
        return run_query(
            query=self.queries.get_all_enrollments_query(),
            params={
                'enterprise_customer_uuid': enterprise_customer_uuid,
                'start_date': start_date,
                'end_date': end_date,
                'limit': limit,
                'offset': offset,
            },
            as_dict=True,
        )

    @cache_it()
    def get_enrollment_date_range(self, enterprise_customer_uuid: UUID):
        """
        Get the enrollment date range for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID | str): The UUID of the enterprise customer.

        Returns:
            (tuple<date, date>): The minimum and maximum enrollment dates.
        """
        results = run_query(
            query=self.queries.get_enrollment_date_range_query(),
            params={'enterprise_customer_uuid': enterprise_customer_uuid}
        )
        if not results:
            return None, None
        min_date, max_date = results[0]

        # We should return date objects, not datetime objects
        # This is done to counter cases where database values are datetime objects.
        if min_date and isinstance(min_date, datetime):
            min_date = min_date.date()
        if max_date and isinstance(max_date, datetime):
            max_date = max_date.date()

        return min_date, max_date

    @cache_it()
    def get_enrollment_and_course_count(self, enterprise_customer_uuid: UUID, start_date: date, end_date: date):
        """
        Get the enrollment and course count for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            (tuple<int, int>): The enrollment and course count.
        """
        results = run_query(
            query=self.queries.get_enrollment_and_course_count_query(),
            params={
                'enterprise_customer_uuid': enterprise_customer_uuid,
                'start_date': start_date,
                'end_date': end_date,
            }
        )
        if not results:
            return 0, 0
        return tuple(results[0])

    @cache_it()
    def get_completion_count(self, enterprise_customer_uuid: UUID, start_date: date, end_date: date):
        """
        Get the completion count for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            int: The completion count.
        """
        results = run_query(
            query=self.queries.get_completion_count_query(),
            params={
                'enterprise_customer_uuid': enterprise_customer_uuid,
                'start_date': start_date,
                'end_date': end_date,
            }
        )
        if not results:
            return 0

        return int(results[0][0] or 0)

    @cache_it()
    def get_top_courses_by_enrollments(self, enterprise_customer_uuid: UUID, start_date: date, end_date: date):
        """
        Get the top courses enrollments for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            list<dict>: A list of dictionaries containing the course key, course_title and enrollment count.
        """
        return run_query(
            query=self.queries.get_top_courses_by_enrollments_query(),
            params={
                'enterprise_customer_uuid': enterprise_customer_uuid,
                'start_date': start_date,
                'end_date': end_date,
            },
            as_dict=True,
        )

    @cache_it()
    def get_top_subjects_by_enrollments(self, enterprise_customer_uuid: UUID, start_date: date, end_date: date):
        """
        Get the top subjects by enrollments for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            list<dict>: A list of dictionaries containing the subject and enrollment count.
        """
        return run_query(
            query=self.queries.get_top_subjects_by_enrollments_query(),
            params={
                'enterprise_customer_uuid': enterprise_customer_uuid,
                'start_date': start_date,
                'end_date': end_date,
            },
            as_dict=True,
        )

    @cache_it()
    def get_enrolment_time_series_data(self, enterprise_customer_uuid: UUID, start_date: date, end_date: date):
        """
        Get the enrollment time series data for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            list<dict>: A list of dictionaries containing the date and enrollment count.
        """
        return run_query(
            query=self.queries.get_enrolment_time_series_data_query(),
            params={
                'enterprise_customer_uuid': enterprise_customer_uuid,
                'start_date': start_date,
                'end_date': end_date,
            },
            as_dict=True,
        )

    @cache_it()
    def get_all_completions(
            self, enterprise_customer_uuid: UUID, start_date: date, end_date: date, limit: int, offset: int
    ):
        """
        Get all completions for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.
            limit (int): The maximum number of records to return.
            offset (int): The number of records to skip.

        Returns:
            list<dict>: A list of dictionaries containing the completions data.
        """
        return run_query(
            query=self.queries.get_all_completions_query(),
            params={
                'enterprise_customer_uuid': enterprise_customer_uuid,
                'start_date': start_date,
                'end_date': end_date,
                'limit': limit,
                'offset': offset,
            },
            as_dict=True,
        )

    @cache_it()
    def get_top_courses_by_completions(self, enterprise_customer_uuid: UUID, start_date: date, end_date: date):
        """
        Get the top courses by completion for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            list<dict>: A list of dictionaries containing the course key, course_title and completion count.
        """
        return run_query(
            query=self.queries.get_top_courses_by_completions_query(),
            params={
                'enterprise_customer_uuid': enterprise_customer_uuid,
                'start_date': start_date,
                'end_date': end_date,
            },
            as_dict=True,
        )

    @cache_it()
    def get_top_subjects_by_completions(self, enterprise_customer_uuid: UUID, start_date: date, end_date: date):
        """
        Get the top subjects by completions for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            list<dict>: A list of dictionaries containing the subject and completion count.
        """
        return run_query(
            query=self.queries.get_top_subjects_by_completions_query(),
            params={
                'enterprise_customer_uuid': enterprise_customer_uuid,
                'start_date': start_date,
                'end_date': end_date,
            },
            as_dict=True,
        )

    @cache_it()
    def get_completions_time_series_data(self, enterprise_customer_uuid: UUID, start_date: date, end_date: date):
        """
        Get the completions time series data for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            list<dict>: A list of dictionaries containing the date and completion count.
        """
        return run_query(
            query=self.queries.get_completions_time_series_data_query(),
            params={
                'enterprise_customer_uuid': enterprise_customer_uuid,
                'start_date': start_date,
                'end_date': end_date,
            },
            as_dict=True,
        )



#!/usr/bin/python3
"""
Sends an Enterprise Customer's data file to a configured destination.
"""


import argparse
import datetime
import logging
import os
import re
import sys

import pytz

from enterprise_reporting.clients.enterprise import EnterpriseAPIClient
from enterprise_reporting.reporter import EnterpriseReportSender
from enterprise_reporting.utils import is_current_time_in_schedule

logging.basicConfig(level=logging.INFO)
LOGGER = logging.getLogger(__name__)
DATA_TYPES = ['progress', 'progress_v2', 'progress_v3', 'catalog', 'grade', 'course_structure', 'completion', 'engagement']


def send_data(config):
    """
    Send data report to each enterprise.

    Args:
        config
    """
    enterprise_customer_name = config['enterprise_customer']['name']
    LOGGER.info(f'Kicking off job to send report for {enterprise_customer_name}')

    error_raised = False
    try:
        reporter = EnterpriseReportSender.create(config)
        reporter.send_enterprise_report()
    except Exception:  # pylint: disable=broad-except
        error_raised = True
        LOGGER.exception(f'Data report failed to send for {enterprise_customer_name}')

    cleanup_files(config['enterprise_customer']['uuid'])
    LOGGER.info(f'Finished job to send report for {enterprise_customer_name}')

    return error_raised


def cleanup_files(enterprise_id):
    """
    Clean up any files created by sending the enterprise report.
    """
    directory = EnterpriseReportSender.FILE_WRITE_DIRECTORY
    pattern = fr'{enterprise_id}'
    for f in os.listdir(directory):
        if re.search(pattern, f):
            os.remove(os.path.join(directory, f))


def should_deliver_report(args, reporting_config, current_est_time):
    """Given CLI arguments and the reporting configuration, determine if delivery should happen."""
    valid_data_type = reporting_config['data_type'] in (args.data_type or DATA_TYPES)
    enterprise_customer_specified = bool(args.enterprise_customer)

    meets_schedule_requirement = is_current_time_in_schedule(
        current_est_time,
        reporting_config['frequency'],
        reporting_config['hour_of_day'],
        reporting_config['day_of_month'],
        reporting_config['day_of_week']
    )
    LOGGER.info("Report Delivery Logic. Active: [%s], ValidDataType: [%s], MeetSchedule: [%s], EnterpriseInArgs: [%s]",
        reporting_config['active'],
        valid_data_type,
        meets_schedule_requirement,
        enterprise_customer_specified,
    )

    return reporting_config['active'] and \
           valid_data_type and \
           (enterprise_customer_specified or meets_schedule_requirement)


def process_reports():
    """
    Process and send reports based on the arguments passed.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('-e', '--enterprise-customer', required=False, type=str,
                        help="Enterprise Customer's UUID. If specified, data delivery is forced.")
    parser.add_argument('-d', '--data-type', required=False, type=str, choices=DATA_TYPES,
                        help="Data type. If specified, only this type of data for the customer(s) will be sent, "
                             "whether forced or not.")
    parser.add_argument('--page-size', required=False, type=int, default=1000,
                        help="The page size to use to retrieve data that comes in a paginated response.")
    args = parser.parse_args()

    enterprise_api_client = EnterpriseAPIClient()
    if args.enterprise_customer:
        reporting_configs = enterprise_api_client.get_enterprise_reporting_configs(args.enterprise_customer)
    else:
        reporting_configs = enterprise_api_client.get_all_enterprise_reporting_configs()

    if args.enterprise_customer and not (reporting_configs and reporting_configs['results']):
        LOGGER.error(f'The enterprise {args.enterprise_customer} does not have a reporting configuration.')
        sys.exit(1)

    # We are defining the current est time globally because we want the current time for a job
    # to remain same thoughout the job. This ensures that a single report is not processed multiple times. 
    # See this comment for more details: https://2u-internal.atlassian.net/browse/ENT-9954?focusedCommentId=5356815
    est_timezone = pytz.timezone('US/Eastern')
    current_est_time = datetime.datetime.now(est_timezone)

    error_raised = False
    for reporting_config in reporting_configs['results']:
        LOGGER.info('Checking if {}\'s reporting config for {} data in {} format is ready for processing'.format(
            reporting_config['enterprise_customer']['name'],
            reporting_config['data_type'],
            reporting_config['report_type'],
        ))

        if should_deliver_report(args, reporting_config, current_est_time):
            if send_data(reporting_config):
                error_raised = True
        else:
            LOGGER.info('Not ready -- skipping this report.')

    if error_raised:
        LOGGER.error(
            'One or more reports in this job were not successfully sent to customers. '
            'Please check these jenkins logs for more information.'
        )
        sys.exit(1)

if __name__ == "__main__":
    process_reports()
    sys.exit(0)



"""
Utility functions for Enterprise Data app.
"""
import hashlib
import random
import time
from contextlib import contextmanager
from datetime import timedelta
from functools import wraps
from logging import getLogger

LOGGER = getLogger(__name__)


def get_cache_key(**kwargs):
    """
    Get MD5 encoded cache key for given arguments.

    Here is the format of key before MD5 encryption.
        key1:value1__key2:value2 ...

    Example:
        >>> get_cache_key(site_domain="example.com", resource="enterprise_customer_users")
        # Here is key format for above call
        # "site_domain:example.com__resource:enterprise_customer_users"
        a54349175618ff1659dee0978e3149ca

    Arguments:
        **kwargs: Keyword arguments that need to be present in cache key.

    Returns:
         An MD5 encoded key uniquely identified by the key word arguments.
    """
    key = '__'.join([f'{item}:{value}' for item, value in kwargs.items()])

    return hashlib.md5(key.encode('utf-8')).hexdigest()


def get_unique_id():
    """
    Return a unique 32 bit integer.
    """
    return random.getrandbits(32)


def subtract_one_month(original_date):
    """
    Return a date exactly one month prior to the passed in date.
    """
    one_day = timedelta(days=1)
    one_month_earlier = original_date - one_day
    while one_month_earlier.month == original_date.month or one_month_earlier.day > original_date.day:
        one_month_earlier -= one_day
    return one_month_earlier


def timeit(func):
    """
    Measure time taken by a function.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        LOGGER.info(f'Time taken by {func.__name__}: {end - start} seconds')
        return result

    return wrapper


@contextmanager
def timer(prefix):
    """
    Context manager to measure the time taken by a block of code.

    Arguments:
        prefix (str): The prefix to print in the log.
    """
    start = time.time()
    yield
    difference = time.time() - start
    LOGGER.info(f"TIMER:: {prefix} took {difference:.20f} seconds")


def primary_subject_truncate(x):
    """
    Truncate primary subject to a few categories.
    """
    if x in [
        "business-management",
        "computer-science",
        "data-analysis-statistics",
        "engineering",
        "communication",
    ]:
        return x
    else:
        return "other"


def find_first(iterable, condition):
    """
    Find the first item in an iterable that satisfies the condition.

    Arguments:
        iterable (iterable): The iterable to search.
        condition (function): The condition to satisfy.

    Returns:
        The first item that satisfies the condition, or None if no item satisfies the condition.
    """
    try:
        return next(item for item in iterable if condition(item))
    except StopIteration:
        return None


def calculate_percentage_difference(first, second):
    """
    Calculate the percentage difference between two numbers.

    It will calculate the percentage difference between the two numbers using the formula:
    ((second - first) / (first + second)/2) * 100

    Arguments:
        first (float): The first number.
        second (float): The second number.

    Returns:
        float: The percentage difference between the two numbers.
    """
    if first == 0 and second == 0:
        return 0
    return ((second - first) / ((first + second) / 2)) * 100



"""
Tests for the `enterprise-data` models module.
"""


import unittest

import ddt
from pytest import mark

from enterprise_data.tests.test_utils import (
    EnterpriseEnrollmentFactory,
    EnterpriseGroupMembershipFactory,
    EnterpriseOfferFactory,
    EnterpriseSubsidyBudgetFactory,
    EnterpriseUserFactory,
)


@mark.django_db
@ddt.ddt
class TestEnterpriseEnrollment(unittest.TestCase):
    """
    Tests for Enterprise Enrollment model
    """

    def setUp(self):
        enterprise_user = EnterpriseUserFactory(enterprise_user_id=1234)
        course_id = 'course-v1:edX+DemoX+DemoCourse'

        self.enrollment = EnterpriseEnrollmentFactory(
            enterprise_user=enterprise_user,
            course_id=course_id
        )
        super().setUp()

    @ddt.data(str, repr)
    def test_string_conversion(self, method):
        """
        Test conversion to string.
        """
        expected_str = '<Enterprise Enrollment for user 1234 in course-v1:edX+DemoX+DemoCourse>'
        assert expected_str == method(self.enrollment)


@mark.django_db
@ddt.ddt
class TestEnterpriseOffer(unittest.TestCase):
    """
    Tests for Enterprise Offer model
    """

    def setUp(self):
        self.enterprise_offer = EnterpriseOfferFactory(
            enterprise_customer_uuid='ee5e6b3a-069a-4947-bb8d-d2dbc323396c',
            enterprise_name='test-enterprise'
        )
        super().setUp()

    @ddt.data(str, repr)
    def test_string_conversion(self, method):
        """
        Test conversion to string.
        """
        expected_str = (f'<Enterprise Offer {self.enterprise_offer.offer_id} '
                        f'for {self.enterprise_offer.enterprise_name}>')
        assert expected_str == method(self.enterprise_offer)


@mark.django_db
@ddt.ddt
class TestEnterpriseSubsidyBudget(unittest.TestCase):
    """
    Tests for Enterprise Subsidy Budget model
    """

    def setUp(self):
        self.enterprise_subsidy_budget = EnterpriseSubsidyBudgetFactory(
            enterprise_customer_uuid='ee5e6b3a-069a-4947-bb8d-d2dbc323396c',
            enterprise_customer_name='test-enterprise'
        )
        super().setUp()

    @ddt.data(str, repr)
    def test_string_conversion(self, method):
        """
        Test conversion to string.
        """
        expected_str = (f'<Enterprise Budget {self.enterprise_subsidy_budget.subsidy_access_policy_uuid} '
                        f'for Subsidy {self.enterprise_subsidy_budget.subsidy_uuid} '
                        f'for {self.enterprise_subsidy_budget.enterprise_customer_name}>')
        assert expected_str == method(self.enterprise_subsidy_budget)


@mark.django_db
@ddt.ddt
class TestEnterpriseGroupMembership(unittest.TestCase):
    """
    Tests for Enterprise Group Membership model
    """

    def setUp(self):
        self.enterprise_group_membership = EnterpriseGroupMembershipFactory(
            enterprise_customer_id='ee5e6b3a-069a-4947-bb8d-d2dbc323396c',
            enterprise_group_name='Test Group',
            enterprise_group_uuid='ee5e6b3a-069a-4947-bb8d-d2dbc323396d',
        )
        super().setUp()

    @ddt.data(str, repr)
    def test_string_conversion(self, method):
        """
        Test conversion to string.
        """
        expected_str = ('<Enterprise Group Membership: Test Group '
                        '(Group UUID: ee5e6b3a-069a-4947-bb8d-d2dbc323396d) '
                        'for Customer ID ee5e6b3a-069a-4947-bb8d-d2dbc323396c>')
        assert expected_str == method(self.enterprise_group_membership)


@mark.django_db
@ddt.ddt
class TestEnterpriseUser(unittest.TestCase):
    """
    Tests for Enterprise User model
    """
    def setUp(self):
        enterprise_id = 'ee5e6b3a-069a-4947-bb8d-d2dbc323396c'
        lms_user_id = 1234
        self.user = EnterpriseUserFactory(lms_user_id=lms_user_id, enterprise_id=enterprise_id)
        super().setUp()

    @ddt.data(str, repr)
    def test_string_conversion(self, method):
        """
        Test conversion to string.
        """
        expected_str = '<Enterprise User 1234 in ee5e6b3a-069a-4947-bb8d-d2dbc323396c>'
        assert expected_str == method(self.user)



# Generated by Django 4.2.15 on 2025-02-20 08:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise_data', '0046_enterprisegroupmembership'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterpriseexecedlcmoduleperformance',
            name='avg_after_lo_score',
            field=models.DecimalField(decimal_places=6, max_digits=38, null=True),
        ),
        migrations.AddField(
            model_name='enterpriseexecedlcmoduleperformance',
            name='avg_before_lo_score',
            field=models.DecimalField(decimal_places=6, max_digits=38, null=True),
        ),
        migrations.AddField(
            model_name='enterpriseexecedlcmoduleperformance',
            name='question_name',
            field=models.CharField(max_length=500, null=True),
        ),
    ]



"""
External Resource Link Report Generation Code.
"""


import logging
import operator
import os
import re
import sys
from collections import Counter
from datetime import date
from urllib.parse import urlparse

from py2neo import Graph

from enterprise_reporting.utils import send_email_with_attachment

logging.basicConfig(level=logging.INFO)
LOGGER = logging.getLogger(__name__)

AGGREGATE_REPORT_CSV_HEADER_ROW = 'Course Key,Course Title,Partner,External Domain,Count\n'


def create_csv_string(processed_results, header_row, additional_columns):
    """
    generates a string suitable to be written as a csv file. it will build
    upon the first string handed in via header_row

    processed_results - a dictionary with course keys as keys and a value that
                        also is a dict, containing keys for course_title and
                        organization (this will constitute the first 3 columns)
    header_row - a string suitable to be written to a csv file
               - must end with a newline
    additional_columns - a callable that takes a course data dict that can
                         return a csv string for any additional columns

    Returns (unicode) string
    """
    for course_key, data in list(processed_results.items()):
        header_row += '{},"{}",{},{}\n'.format(
            course_key,
            data['course_title'],
            data['organization'],
            additional_columns(data),
        )
    return header_row


def create_columns_for_aggregate_report(data):
    """
    Creates a csv string for additional columns in report
    """
    urls_sorted_by_counts = sorted(
        list(data['domain_count'].items()),
        key=operator.itemgetter(1),
        reverse=True
    )
    stringified_urls_and_counts = [
        f'{url},{count}'
        for url, count in urls_sorted_by_counts
    ]
    return '\n,,,'.join(stringified_urls_and_counts)


def gather_links_from_html(html_string):
    """
    Takes some html blob as a string and extracts any external links, and
    returns them as a set
    """
    pattern = 'https?://.*?[" <]'

    links = []
    for link in re.findall(pattern, html_string,):
        link = link[0:-1].strip()
        if (link.lower().endswith('.png"') or
            link.lower().endswith('.jpg"') or
            link.lower().endswith('.jpeg"') or
            link.lower().endswith('.gif"') or
            '.edx.org' in link):
            continue

        # Want to verify the link captured is a proper url
        # If not, toss it out and throw a log message
        try:
            urlparse(link)
        except ValueError:
            LOGGER.warning(
                "Unparsable URL found. Not including in report: %s" % link
            )
            continue

        links.append(link)
    return links


def process_coursegraph_results(raw_results):
    """
    Takes the data from a coursegraph query

    Returns a dict with course keys as the key and dict data about that course
    as value
    """
    processed_results = {}
    for entry in raw_results:
        course_key = entry['h.course_key']
        # Only want new style course keys to exclude archived courses
        if not course_key.startswith('course-'):
            continue

        external_links = gather_links_from_html(entry['h.data'])

        if not external_links:
            continue

        # Create entry in processed_results and keep external_links up to date
        if course_key not in processed_results:
            processed_results[course_key] = {
                'course_title': entry['course_title'],
                'organization': entry['organization'],
                'external_links': set(external_links),
                'domain_count': {},
            }
        else:
            processed_results[course_key]['external_links'].update(external_links)

        # Now use external links to create the domains and counts
        domains = [
            '{uri.scheme}://{uri.netloc}'.format(uri=urlparse(link))
            for link in external_links
        ]
        # calculate the unique counts for all the urls
        domains_with_counts = dict(Counter(domains))

        for domain, count in list(domains_with_counts.items()):
            if domain in processed_results[course_key]['domain_count']:
                processed_results[course_key]['domain_count'][domain] += count
            else:
                processed_results[course_key]['domain_count'][domain] = count

    return processed_results


def query_coursegraph():
    """
    Calls coursegraph with cypher query and returns query data

    The data is a list of dicts where each dict contains keys that correspond
    to what is being returned in the query
    """
    graph = Graph(
        bolt=True,
        http_port=os.environ.get('COURSEGRAPH_PORT'),
        host=os.environ.get('COURSEGRAPH_HOST'),
        secure=True,
    )
    query = '''MATCH
                (c:course)-[:PARENT_OF*]->(h:html) 
              WHERE (
                h.data CONTAINS 'https://'
                OR
                h.data CONTAINS 'http://'
              )
              RETURN
                c.display_name as course_title,
                c.org as organization,
                h.course_key, 
                h.data'''
    results = graph.run(query)
    return results.data()


def split_up_results(processed_results_part1):
    """
    Splits up a dictionary of processed_results into a list of
    two smaller dictionaries.
    """
    # NOTE: we are splitting this into 2 dicts because it is 1) easy to do
    # and 2) we have about 8 MB of overhead for newly added strings to this
    # report before it becomes a problem again... which should not be for
    # a veeeery long time.
    #
    # If emails fail due to size in the future, implement a solution that
    # dynamically chooses chunk size and email quantity based on size of
    # string data included in each of the dictionary entries being
    # iterated over.
    processed_results_part2 = {}

    original_len = len(processed_results_part1)
    while len(processed_results_part1) > (original_len / 2):
        k, v = processed_results_part1.popitem()
        processed_results_part2[k] = v

    return [processed_results_part1, processed_results_part2]


def generate_and_email_report():
    """
    Generates a report an sends it as an email with an attachment
    """

    from_email = os.environ.get('SEND_EMAIL_FROM')
    today = str(date.today())
    subject = 'External Resource Link Report'
    body = '''Dear Customer Success,
Find attached a file containing course keys and their respective
external resource links.

If you have any questions/concerns with the report, please ask the
Enterprise Team!

Sincerely,
The Enterprise Team'''

    LOGGER.info("Querying Course Graph DB...")
    raw_results = query_coursegraph()

    # Results are too large to send in 1 email (~10MB is the limit) so
    # we split up the results into two parts
    processed_results = process_coursegraph_results(raw_results)
    result_dicts = split_up_results(processed_results)

    for index, result in enumerate(result_dicts):
        readable_number = index + 1

        LOGGER.info(
            "Generating aggregate external links spreadsheet part %s..." % readable_number
        )
        aggregate_report_data = create_csv_string(
            result,
            AGGREGATE_REPORT_CSV_HEADER_ROW,
            create_columns_for_aggregate_report
        )
        filename = 'external-resource-domain-report-part{}-{}.csv'.format(
            readable_number,
            today,
        )
        attachment_data = {filename: aggregate_report_data.encode('utf-8')}

        LOGGER.info(
            "Emailing aggregate spreadsheet part %s..." % readable_number
        )
        send_email_with_attachment(
            subject,
            body,
            from_email,
            TO_EMAILS.split(','),
            attachment_data
        )


if __name__ == '__main__':
    TO_EMAILS = sys.argv[1]
    generate_and_email_report()



"""
URL definitions for enterprise data api v1.
"""


from rest_framework.routers import DefaultRouter

from django.urls import re_path

from enterprise_data.api.v1.views import enterprise_admin as enterprise_admin_views
from enterprise_data.api.v1.views import enterprise_learner as enterprise_learner_views
from enterprise_data.api.v1.views import enterprise_offers as enterprise_offers_views
from enterprise_data.api.v1.views.analytics_completions import AdvanceAnalyticsCompletionsView
from enterprise_data.api.v1.views.analytics_engagements import AdvanceAnalyticsEngagementView
from enterprise_data.api.v1.views.analytics_enrollments import AdvanceAnalyticsEnrollmentsView
from enterprise_data.api.v1.views.analytics_leaderboard import AdvanceAnalyticsLeaderboardView
from enterprise_data.constants import UUID4_REGEX

app_name = 'enterprise_data_api_v1'

router = DefaultRouter()
router.register(
    r'enterprise/(?P<enterprise_id>.+)/enrollments',
    enterprise_learner_views.EnterpriseLearnerEnrollmentViewSet,
    'enterprise-learner-enrollment',
)
router.register(
    r'enterprise/(?P<enterprise_id>.+)/offers',
    enterprise_offers_views.EnterpriseOfferViewSet,
    'enterprise-offers',
)
router.register(
    r'enterprise/(?P<enterprise_id>.+)/users',
    enterprise_learner_views.EnterpriseLearnerViewSet,
    'enterprise-learner',
)
router.register(
    r'enterprise/(?P<enterprise_id>.+)/learner_completed_courses',
    enterprise_learner_views.EnterpriseLearnerCompletedCoursesViewSet,
    'enterprise-learner-completed-courses',
)
router.register(
    r'enterprise/(?P<enterprise_id>.+)/module-performance',
    enterprise_admin_views.EnterpriseExecEdLCModulePerformanceViewSet,
    'enterprise-admin-module-performance',
)

urlpatterns = [
    re_path(
        fr'^admin/insights/(?P<enterprise_id>{UUID4_REGEX})$',
        enterprise_admin_views.EnterpriseAdminInsightsView.as_view(),
        name='enterprise-admin-insights'
    ),
    re_path(
        fr'^admin/analytics/(?P<enterprise_id>{UUID4_REGEX})$',
        enterprise_admin_views.EnterpriseAdminAnalyticsAggregatesView.as_view(),
        name='enterprise-admin-analytics-aggregates'
    ),
    re_path(
        fr'^admin/analytics/(?P<enterprise_uuid>{UUID4_REGEX})/leaderboard$',
        AdvanceAnalyticsLeaderboardView.as_view({'get': 'list'}),
        name='enterprise-admin-analytics-leaderboard-list'
    ),
    re_path(
        fr'^admin/analytics/(?P<enterprise_uuid>{UUID4_REGEX})/enrollments/stats$',
        AdvanceAnalyticsEnrollmentsView.as_view({'get': 'stats'}),
        name='enterprise-admin-analytics-enrollments-stats'
    ),
    re_path(
        fr'^admin/analytics/(?P<enterprise_uuid>{UUID4_REGEX})/enrollments$',
        AdvanceAnalyticsEnrollmentsView.as_view({'get': 'list'}),
        name='enterprise-admin-analytics-enrollments'
    ),
    re_path(
        fr'^admin/analytics/(?P<enterprise_uuid>{UUID4_REGEX})/completions/stats$',
        AdvanceAnalyticsCompletionsView.as_view({'get': 'stats'}),
        name='enterprise-admin-analytics-completions-stats'
    ),
    re_path(
        fr'^admin/analytics/(?P<enterprise_uuid>{UUID4_REGEX})/completions$',
        AdvanceAnalyticsCompletionsView.as_view({'get': 'list'}),
        name='enterprise-admin-analytics-completions'
    ),
    re_path(
        fr'^admin/analytics/(?P<enterprise_uuid>{UUID4_REGEX})/engagements/stats$',
        AdvanceAnalyticsEngagementView.as_view({'get': 'stats'}),
        name='enterprise-admin-analytics-engagements-stats'
    ),
    re_path(
        fr'^admin/analytics/(?P<enterprise_uuid>{UUID4_REGEX})/engagements$',
        AdvanceAnalyticsEngagementView.as_view({'get': 'list'}),
        name='enterprise-admin-analytics-engagements'
    ),
    re_path(
        fr'^admin/analytics/(?P<enterprise_id>{UUID4_REGEX})/skills/stats',
        enterprise_admin_views.EnterpriseAdminAnalyticsSkillsView.as_view(),
        name='enterprise-admin-analytics-skills'
    ),
    re_path(
        fr'^enterprise/(?P<enterprise_uuid>{UUID4_REGEX})/budgets',
        enterprise_admin_views.EnterpriseBudgetView.as_view(),
        name='enterprise-budgets'
    ),
    re_path(
        fr'^enterprise/(?P<enterprise_uuid>{UUID4_REGEX})/groups',
        enterprise_admin_views.EnterpriseGroupMembershipView.as_view(),
        name='enterprise-groups'
    ),
]

urlpatterns += router.urls



"""
Constants for enterprise data app.
"""


ENTERPRISE_DATA_ADMIN_ROLE = 'enterprise_data_admin'
SYSTEM_ENTERPRISE_ADMIN_ROLE = 'enterprise_admin'
SYSTEM_ENTERPRISE_OPERATOR_ROLE = 'enterprise_openedx_operator'

# this constant should not be used anymore as it is marked for removal
# switch is used to enable/disable role based access control
ROLE_BASED_ACCESS_CONTROL_SWITCH = 'role_based_access_control'

# context to give access to all resources
ALL_ACCESS_CONTEXT = '*'
